import ctypes
import time
import traceback
from ctypes import wintypes

import win32con
import win32gui

try:
    from PIL import Image
except:
    print("Pillow lib is not installed")


class ICONINFO(ctypes.Structure):
    _fields_ = [
        ("fIcon", ctypes.c_bool),  # Specifies whether this structure defines an icon or a cursor
        ("xHotspot", ctypes.c_ulong),  # The x-coordinate of a cursor's hot spot
        ("yHotspot", ctypes.c_ulong),  # The y-coordinate of a cursor's hot spot
        ("hbmMask", ctypes.c_void_p),  # Handle to the icon/cursor mask bitmap
        ("hbmColor", ctypes.c_void_p),  # Handle to the icon/cursor color bitmap
    ]


def _get_icon_info(h_icon):
    icon_info = ICONINFO()
    ctypes.windll.user32.GetIconInfo(h_icon, ctypes.byref(icon_info))
    return icon_info


def _get_bitmap_bits(hbm, width, height):
    """
    Retrieve raw bitmap data using ctypes and the GetBitmapBits API.
    """
    buffer_len = width * height * 4  # 4 bytes per pixel (RGBA)
    buffer = ctypes.create_string_buffer(buffer_len)

    hbm = wintypes.HBITMAP(hbm)

    # Call GetBitmapBits from the GDI API
    success = ctypes.windll.gdi32.GetBitmapBits(hbm, buffer_len, buffer)
    if not success:
        raise RuntimeError("Failed to retrieve bitmap bits.")

    return buffer.raw


def _get_window_icon(hwnd) -> "Image":
    h_icon = win32gui.SendMessage(hwnd, win32con.WM_GETICON, win32con.ICON_BIG, 0)
    if not h_icon:
        h_icon = win32gui.SendMessage(hwnd, win32con.WM_GETICON, win32con.ICON_SMALL, 0)
    if not h_icon:
        # In most cases this thing works. .Returnes 40x40 image
        h_icon = ctypes.windll.user32.GetClassLongPtrW(hwnd, win32con.GCL_HICON)
    if not h_icon:
        raise ValueError("No icon found for this window.")

    # Get ICONINFO structure
    icon_info = _get_icon_info(h_icon)

    # Extract color and mask bitmaps
    hbm_color = icon_info.hbmColor
    hbm_mask = icon_info.hbmMask

    # Get dimensions from the color bitmap
    bmp_info = win32gui.GetObject(hbm_color)
    width, height = bmp_info.bmWidth, bmp_info.bmHeight

    # Create a device context for drawing
    hdc = win32gui.GetDC(0)
    hdc_mem = win32gui.CreateCompatibleDC(hdc)
    win32gui.SelectObject(hdc_mem, hbm_color)

    # Extract raw pixel data
    bmp_bits = _get_bitmap_bits(hbm_color, width, height)
    img = Image.frombuffer("RGBA", (width, height), bmp_bits, "raw", "BGRA", 0, 1)

    # Release resources
    win32gui.DeleteObject(hbm_color)
    win32gui.DeleteObject(hbm_mask)
    win32gui.DeleteDC(hdc_mem)
    win32gui.ReleaseDC(0, hdc)

    return img.resize((40, 40))


# import win32gui
# import win32ui
# import win32con
# import time
# from PIL import Image
# import os
# import ctypes


# def get_window_icon(hwnd):
#     # NEVER SEEN THIS ICON BEEN FOUND
#     h_icon = win32gui.SendMessage(hwnd, win32con.WM_GETICON, win32con.ICON_BIG, 0)
#     if not h_icon:
#         # NEVER SEEN THIS ICON BEEN FOUND
#         h_icon = win32gui.SendMessage(hwnd, win32con.WM_GETICON, win32con.ICON_SMALL, 0)
#     if not h_icon:
#         # ALWAYS FOUND THIS OR NONE
#         h_icon = ctypes.windll.user32.GetClassLongPtrW(hwnd, win32con.GCL_HICON)
#     if not h_icon:
#         raise ValueError("No icon found for this window.")

#     hdc = win32ui.CreateDCFromHandle(win32gui.GetDC(0))
#     hdc_mem = hdc.CreateCompatibleDC()
#     bmp = win32ui.CreateBitmap()
#     bmp.CreateCompatibleBitmap(hdc, 256, 256)  # Allocate enough space for a large icon
#     hdc_mem.SelectObject(bmp)

#     # Draw the icon into the bitmap
#     win32gui.DrawIconEx(hdc_mem.GetHandleOutput(), 0, 0, h_icon, 256, 256, 0, None, win32con.DI_NORMAL)

#     # Save the bitmap as an image file
#     bmp_info = bmp.GetInfo()
#     bmp_data = bmp.GetBitmapBits(True)
#     win32gui.DeleteObject(bmp.GetHandle())
#     hdc_mem.DeleteDC()
#     hdc.DeleteDC()
#     image = Image.frombuffer("RGB", (bmp_info["bmWidth"], bmp_info["bmHeight"]), bmp_data, "raw", "BGRX", 0, 1)
#     return image


# def main():

#     from winytils.windows import Windows

#     # Find the target window
#     # hwnd = win32gui.GetForegroundWindow()
#     window = Windows.foregrond()
#     if not window:
#         print("Window not found!")
#         return

#     if window.is_uwp():
#         child = window.uwp_child
#         if child:
#             window = child

#     try:
#         img = get_window_icon(window.hwnd)
#         img.save("i.png")
#     except Exception as e:
#         print(f"Error: {e}")


# if __name__ == "__main__":
#     time.sleep(3)
#     main()



import threading
import time
from collections import defaultdict
from datetime import datetime
from enum import Enum
from typing import Callable

import psutil
import win32api
import win32con
import win32gui
import win32ts


class SessionEvent(Enum):
    ANY = 0
    CHANGE = 0x2B1
    CONSOLE_CONNECT = 0x1
    CONSOLE_DISCONNECT = 0x2
    REMOTE_CONNECT = 0x3
    REMOTE_DISCONNECT = 0x4
    SESSION_LOGON = 0x5
    SESSION_LOGOFF = 0x6
    SESSION_LOCK = 0x7
    SESSION_UNLOCK = 0x8
    SESSION_REMOTE_CONTROL = 0x9


class WorkstationMonitor:
    CLASS_NAME = "WorkstationMonitor"
    WINDOW_TITLE = "Workstation Event Monitor"

    def __init__(self):
        self.window_handle = None
        self.event_handlers = defaultdict(list)
        self._register_listener()

    def _register_listener(self):
        """Registers the listener with Win32 GUI to listen to session events."""
        wc = win32gui.WNDCLASS()
        wc.hInstance = win32api.GetModuleHandle(None)
        wc.lpszClassName = self.CLASS_NAME
        wc.lpfnWndProc = self._window_procedure  # This is the window procedure

        # Register the window class
        window_class = win32gui.RegisterClass(wc)

        # Create the window with the correct arguments
        self.window_handle = win32gui.CreateWindow(
            window_class,  # Window class
            self.WINDOW_TITLE,  # Window title
            0,  # Style
            0,  # x-coordinate
            0,  # y-coordinate
            win32con.CW_USEDEFAULT,  # Width
            win32con.CW_USEDEFAULT,  # Height
            0,  # Parent window (None, as 0)
            0,  # Menu (None, as 0)
            wc.hInstance,  # Instance handle
            None,  # Additional parameter (None here)
        )

        # Update the window
        win32gui.UpdateWindow(self.window_handle)

        # Register session notifications
        # win32ts.NOTIFY_FOR_THIS_SESSION
        win32ts.WTSRegisterSessionNotification(self.window_handle, win32ts.NOTIFY_FOR_ALL_SESSIONS)

    def listen(self):
        """Start listening for events."""
        win32gui.PumpMessages()

    def stop(self):
        """Stop the listener."""
        win32gui.PostQuitMessage(0)

    def _window_procedure(self, hwnd: int, msg: int, wparam, lparam):
        """Processes window messages and invokes handlers."""
        if msg == SessionEvent.CHANGE.value:
            self._handle_session_change(SessionEvent(wparam), lparam)
        elif msg in [win32con.WM_CLOSE, win32con.WM_DESTROY]:
            win32gui.DestroyWindow(hwnd)
            win32gui.PostQuitMessage(0)
            return 0
        return win32gui.DefWindowProc(hwnd, msg, wparam, lparam)

    def _handle_session_change(self, event: SessionEvent, session_id: int):
        """Handles session change events."""
        for handler in self.event_handlers[event]:
            handler(event, session_id)
        for handler in self.event_handlers[SessionEvent.ANY]:
            handler(event, session_id)

    def register_handler(self, event: SessionEvent, handler: callable):
        """Registers a handler for a specific event."""
        self.event_handlers[event].append(handler)


class MachineStateScanner(threading.Thread):
    # fmt:off
    __slots__ = ("wm", "locked", "unlocked", "on_lock_func", "on_lock_args", "on_lock_kwargs", "on_unlock_func", "on_unlock_args", "on_unlock_kwargs", "_t_lock")
    # fmt:on

    def __init__(self):
        super().__init__()
        self.wm = None
        self.locked: bool = False
        self.unlocked: bool = False
        self.on_lock_func: callable = None
        self.on_lock_args = []
        self.on_lock_kwargs = {}
        self.on_unlock_func: callable = None
        self.on_unlock_args = []
        self.on_unlock_kwargs = {}
        self._t_lock = threading.Lock()

    def run(self):
        """Run method for the thread, used to start the listener."""
        self.wm = WorkstationMonitor()
        self.wm.register_handler(SessionEvent.ANY, self.handler)
        self.wm.listen()  # Do not try listen in thread as it won't work

    def stop(self):
        """Stop the workstation monitor."""
        try:
            self.wm.stop()
        except Exception as e:
            pass

    def on_lock(self, func: callable, *args, **kwargs):
        """Sets the lock handler."""
        with self._t_lock:
            self.on_lock_func = func
            self.on_lock_args = args
            self.on_lock_kwargs = kwargs

    def on_unlock(self, func: callable, *args, **kwargs):
        """Sets the unlock handler."""
        with self._t_lock:
            self.on_unlock_func = func
            self.on_unlock_args = args
            self.on_unlock_kwargs = kwargs

    def handler(self, event, session_id):
        """Main event handler for lock/unlock events."""
        with self._t_lock:
            if event == SessionEvent.SESSION_LOCK:
                self.locked = True
                self.unlocked = False
                if self.on_lock_func:
                    self.on_lock_func(*self.on_lock_args, **self.on_lock_kwargs)
            elif event == SessionEvent.SESSION_UNLOCK:
                self.locked = False
                self.unlocked = True
                if self.on_unlock_func:
                    self.on_unlock_func(*self.on_unlock_args, **self.on_unlock_kwargs)


def watch_lock_unlock(
    thread=None,
    on_lock=lambda: print(f"locked {datetime.now()}"),
    on_unlock=lambda: print(f"unlocked {datetime.now()}"),
    *args,
    **kwargs,
):
    """
    thread: StoppableThread
    """
    ms = MachineStateScanner()
    ms.start()
    try:
        already_locked = False
        already_unlocked = False
        while True:

            if thread and getattr(thread, "is_stopped") and getattr(thread, "is_stopped")():
                raise Exception("Stopped")

            if ms.locked:
                if not already_locked:
                    on_lock(*args, **kwargs)
                    already_locked = True
                    already_unlocked = False
            elif ms.unlocked:
                if not already_unlocked:
                    on_unlock(*args, **kwargs)
                    already_unlocked = True
                    already_locked = False
            time.sleep(0.1)
    except Exception as e:
        ms.stop()
        ms.join()


def _handler(event, session_id):
    """Simple handler for lock/unlock events."""
    if event == SessionEvent.SESSION_LOCK:
        print("Session Locked")
    elif event == SessionEvent.SESSION_UNLOCK:
        print("Session Unlocked")


def _example_1():
    """Example for basic event listening."""
    wm = WorkstationMonitor()
    wm.register_handler(SessionEvent.ANY, handler=_handler)
    wm.listen()


def _example_2():
    """TaskScheduler example with custom lock/unlock handlers."""
    ms = MachineStateScanner()
    ms.on_lock(lambda: print(f"locked {datetime.now()}"))
    ms.on_unlock(lambda: print(f"unlocked {datetime.now()}"))
    ms.start()
    try:
        ms.join()
    except KeyboardInterrupt:
        print("Stopping...")
        ms.stop()
        ms.join()


def get_system_boot_time():
    boot_timestamp = psutil.boot_time()
    boot_dt = datetime.fromtimestamp(boot_timestamp)
    return boot_dt


if __name__ == "__main__":
    _example_2()  # Choose the example you want to run.



import base64
import io
import sys
import time
from typing import Callable, List

import psutil
import pywintypes
import win32api
import win32con
import win32gui
import win32process

from .utils import _get_window_icon

try:
    from PIL import Image
except:
    pass


class Window:
    # This is wrapper for UWP apps
    _UWP_EXE = "ApplicationFrameHost.exe"

    def __init__(self, hwnd: int) -> None:
        self.hwnd = hwnd

    @property
    def title(self) -> str:
        return win32gui.GetWindowText(self.hwnd)

    @property
    def process(self) -> psutil.Process:
        return psutil.Process(self._process_id)

    @property
    def icon(self) -> "Image":
        if "PIL.Image" not in sys.modules:
            raise ImportError("Missing Pillow lib")
        try:
            return _get_window_icon(self.hwnd)
        except:
            return None

    @property
    def icon_base64(self) -> str:
        icon = self.icon
        if not icon:
            return None
        buffer = io.BytesIO()
        icon.save(buffer, format="PNG")
        buffer.seek(0)
        icon_base64 = base64.b64encode(buffer.read()).decode("utf-8")
        return icon_base64

    @property
    def class_name(self) -> str:
        return win32gui.GetClassName(self.hwnd)

    @property
    def children(self) -> List["Window"]:
        """Retrieve all child window handles of the current window."""
        child_windows = []

        def callback(hwnd, windows):
            windows.append(Window(hwnd))
            return True

        win32gui.EnumChildWindows(self.hwnd, callback, child_windows)
        return child_windows

    @property
    def uwp_child(self) -> "Window":
        """Skip UWP"""
        for child in self.children:
            try:
                child_thread, child_pid = win32process.GetWindowThreadProcessId(child.hwnd)
                child_process = psutil.Process(child_pid)
                # Exclude ApplicationFrameHost.exe to avoid recursion
                if child_process.name().lower() != self._UWP_EXE.lower():
                    return child
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue

    def minimize(self, verify_timeout: float = None) -> bool:
        win32gui.ShowWindow(self.hwnd, win32con.SW_MINIMIZE)

        if verify_timeout:
            stop_t = time.time() + 2
            while not self.is_minimized():
                if time.time() > stop_t:
                    return False
                time.sleep(0.01)
            return True
        else:
            return None

    def maximize(self, verify_timeout: float = None) -> bool:
        win32gui.ShowWindow(self.hwnd, win32con.SW_MAXIMIZE)

        if verify_timeout:
            stop_t = time.time() + 2
            while self.is_minimized():
                if time.time() > stop_t:
                    return False
                time.sleep(0.01)
            return True
        else:
            return None

    def restore(self, verify_timeout: float = None) -> bool:
        win32gui.ShowWindow(self.hwnd, win32con.SW_RESTORE)  # SW_SHOWNORMAL

        if verify_timeout:
            stop_t = time.time() + 2
            while self.is_minimized():
                if time.time() > stop_t:
                    return False
                time.sleep(0.01)
            return True
        else:
            return None

    def activate(self):
        win32gui.SetForegroundWindow(self.hwnd)

    def is_uwp(self) -> bool:
        return self.process.name().lower() == self._UWP_EXE.lower()

    def is_minimized(self) -> bool:
        # # SW_SHOWMINIMIZED (2) indicates the window is minimized.
        # placement = win32gui.GetWindowPlacement(self.hwnd) (2, 2, (-25600, -25600), (-1, -1), (40, 0, 1078, 808))
        # return placement[1] == win32con.SW_SHOWMINIMIZED
        return bool(win32gui.IsIconic(self.hwnd))

    def is_maximized(self) -> bool:
        placement = win32gui.GetWindowPlacement(self.hwnd)
        return placement[1] == win32con.SW_SHOWMAXIMIZED

    def is_normal(self) -> bool:
        placement = win32gui.GetWindowPlacement(self.hwnd)
        return placement[1] == win32con.SW_SHOWNORMAL

    def is_visiable(self) -> bool:
        return win32gui.IsWindowVisible(self.hwnd) == 1

    def close(self, verify_timeout: int = None) -> bool:
        try:
            win32gui.PostMessage(self.hwnd, win32con.WM_CLOSE, 0, 0)
        except pywintypes.error as e:
            return False  # Access is denied

        if verify_timeout:
            # Optionally, wait and verify if the window has closed
            # Wait for some time
            stop_t = time.time() + 5
            while not self.is_closed():
                if time.time() > stop_t:
                    return False
                time.sleep(0.01)
            return True
        else:
            return None

    def is_closed(self) -> bool:
        if not win32gui.IsWindow(self.hwnd):
            return True
        else:
            return False

    def kill(self, verify_timeout: float = None) -> bool:
        process = self.process
        process.kill()
        if verify_timeout:
            stop_t = time.time() + verify_timeout
            while process.is_running():
                if time.time() > stop_t:
                    return False
                time.sleep(0.01)
            return True
        else:
            return None

    @property
    def _thread_id(self) -> int:
        thread_id, process_id = win32process.GetWindowThreadProcessId(self.hwnd)
        return thread_id

    @property
    def _process_id(self) -> int:
        thread_id, process_id = win32process.GetWindowThreadProcessId(self.hwnd)
        return process_id

    def __repr__(self) -> str:
        return str({"title": self.title, "exe": self.process.exe(), "class": self.class_name})


class Windows:

    def __init__(self) -> None:
        pass

    @staticmethod
    def filter(
        *,
        has_gui: bool = None,
        opened: bool = None,
        has_title: bool = None,
        by_title: str = None,
        by_exe: str = None,
        by_class_name: str = None,
        custom_filter: Callable = None,
    ) -> List[Window]:
        """
        Z axis from tompost to last
        """
        windows = Windows.get_all()
        # Filter windows that have GUI
        if has_gui == True:
            windows = [w for w in windows if w.is_visiable()]
        elif has_gui == False:
            windows = [w for w in windows if not w.is_visiable()]
        # Filter winows that are maximized(opened)
        if opened == True:
            windows = [w for w in windows if not w.is_minimized()]
        elif opened == False:
            windows = [w for w in windows if w.is_minimized()]
        # Filter windows that have title
        if has_title == True:
            windows = [w for w in windows if w.title]
        elif has_title == False:
            windows = [w for w in windows if not w.title]

        if by_title or by_exe or by_class_name:
            windows = Windows._filter_windows_by(windows, title=by_title, exe=by_exe, class_name=by_class_name)

        if callable(custom_filter):
            windows = custom_filter(windows)

        return windows

    @staticmethod
    def foregrond() -> Window:
        """"""
        # windows = Windows.filter(has_gui=has_gui, opened=opened, has_title=has_title, custom_filter=custom_filter)
        # return windows[0] if windows[:1] else None
        hwnd = win32gui.GetForegroundWindow()
        return Window(hwnd)

    @staticmethod
    def get_all() -> List[Window]:
        """Ordered starting from topmost"""
        windows = []
        for hwnd in Windows._get_hwnds():
            windows.append(Window(hwnd))
        return windows

    @staticmethod
    def _get_hwnds() -> List[int]:
        hwnds = []

        def callback(hwnd, extra):
            hwnds.append(hwnd)
            return True

        win32gui.EnumWindows(callback, None)  # None is extra
        return hwnds

    @staticmethod
    def _filter_windows_by(windows: List[Window], *, title=None, exe=None, class_name=None) -> List[Window]:
        if not (title or exe or class_name):
            raise ValueError("At least one of title, exe, or class_name must be provided")

        found_windows = []
        for w in windows:
            if title is not None and w.title != title:
                continue
            if exe is not None and w.process.name() != exe:
                continue
            if class_name is not None and w.class_name != class_name:
                continue
            found_windows.append(w)

        return found_windows


def windows_filter(windows: List[Window]) -> List[Window]:
    """
    Filters all explorer.exe opened windows(taskbar etc.) except "File Explorer" windows
    """
    filtered_windows = []
    for w in windows:
        # Skip all "explorer.exe" that are not File Explorer
        if w.process.name() == "explorer.exe" and w.class_name not in ["CabinetWClass", "ExplorerWClass"]:
            continue
        else:
            filtered_windows.append(w)
    return filtered_windows


def is_uwp_loaded(w: Window, timeout: float = 5) -> bool:
    """
    Can return False in case if application with no gui is opened so there will be no child [SystemSettings.exe example case]
    In this case it is recommended to do w.minimize() and skip step. So this window should not appear again
    timeout: Timeout to wait for load after that return False. Defaults to 5 sec
    """
    if not w.is_uwp():
        raise Exception("It is not UWP wrapper app")
    stop_t = time.time() + timeout
    child = w.uwp_child
    while not child:
        if time.time() > stop_t:
            return False
        child = w.uwp_child
    return True


def minimize_uwp(w: Window, delay: float = 0.2):
    """Minimum 0.2 sec before child minification"""
    if not w.is_uwp():
        raise Exception("It is not UWP wrapper app")

    uwp_child_win = w.uwp_child
    if not uwp_child_win:
        raise Exception("Missing UWP app child")

    w.minimize()
    time.sleep(delay)
    uwp_child_win.minimize()


def set_window_topmost(hwnd):
    """Set the specified window as topmost."""
    win32gui.SetWindowPos(hwnd, win32con.HWND_TOPMOST, 0, 0, 0, 0, win32con.SWP_NOMOVE | win32con.SWP_NOSIZE)
    print("Window set to topmost.")


def set_window_fullscreen(hwnd):
    """Make the specified window fullscreen."""
    # Get screen dimensions
    screen_width = win32api.GetSystemMetrics(win32con.SM_CXSCREEN)
    screen_height = win32api.GetSystemMetrics(win32con.SM_CYSCREEN)

    style = win32gui.GetWindowLong(hwnd, win32con.GWL_STYLE)

    new_style = style & ~win32con.WS_CAPTION & ~win32con.WS_THICKFRAME

    win32gui.SetWindowLong(hwnd, win32con.GWL_STYLE, new_style)

    flags = win32con.SWP_NOMOVE | win32con.SWP_NOSIZE | win32con.SWP_NOZORDER | win32con.SWP_FRAMECHANGED
    win32gui.SetWindowPos(hwnd, None, 0, 0, 0, 0, flags)


def set_window_overrideredirect(hwnd):
    """Remove title bar, borders, taskbar icon"""
    # Get current extended window styles
    ex_style = win32gui.GetWindowLong(hwnd, win32con.GWL_EXSTYLE)

    # Modify the extended styles
    ex_style &= ~win32con.WS_EX_APPWINDOW  # Remove WS_EX_APPWINDOW
    ex_style |= win32con.WS_EX_TOOLWINDOW  # Add WS_EX_TOOLWINDOW

    # Apply the new extended styles
    win32gui.SetWindowLong(hwnd, win32con.GWL_EXSTYLE, ex_style)

    # Update the window's non-client area to reflect changes
    flags = win32con.SWP_NOMOVE | win32con.SWP_NOSIZE | win32con.SWP_NOZORDER | win32con.SWP_FRAMECHANGED
    win32gui.SetWindowPos(hwnd, None, 0, 0, 0, 0, flags)


def set_window_transparency(hwnd, transparency: int):
    """
    Set the transparency level of a window.
    :param hwnd: Handle to the window.
    :param transparency: Transparency level (0-255). 255 is fully opaque, 0 is fully transparent.
    """
    # Get the current window styles
    ex_style = win32gui.GetWindowLong(hwnd, win32con.GWL_EXSTYLE)

    # Add layered window style if not already set
    if not (ex_style & win32con.WS_EX_LAYERED):
        ex_style |= win32con.WS_EX_LAYERED
        win32gui.SetWindowLong(hwnd, win32con.GWL_EXSTYLE, ex_style)

    # ctypes.windll.user32.SetLayeredWindowAttributes(hwnd, color_key, 0, win32con.LWA_COLORKEY)
    win32gui.SetLayeredWindowAttributes(hwnd, 0, transparency, win32con.LWA_ALPHA)


if __name__ == "__main__":
    # time.sleep(3)
    # windows = Windows.filter(has_gui=True, opened=True, has_title=True, custom_filter=windows_filter)
    windows = Windows.filter(has_gui=True)
    program_manager = Windows.filter(by_title="Program Manager", by_exe="explorer.exe")[0]
    task_switching = Windows.filter(by_title="Task Switching", by_exe="explorer.exe")[0]
    explorer_windows = Windows.filter(by_exe="explorer.exe")
    explorer_windows2 = Windows.filter(has_title=False, by_exe="explorer.exe")

    windows = Windows.filter(by_exe="explorer.exe")
    while True:
        time.sleep(0.3)
        classes = [w.class_name for w in windows]
        print(classes)
        windows = Windows.filter(by_exe="explorer.exe")

        # # print()
        # set_a = set(classes)
        # set_b = set([w.class_name for w in windows])

        # # Find unique items using set difference
        # unique_to_a = set_a - set_b  # Items in A but not in B
        # unique_to_b = set_b - set_a  # Items in B but not in A
        # print(list(unique_to_a) + list(unique_to_b))
        # print(len(windows))
        # print("\n\n")
        # popup_host = Windows.filter(by_title="PopupHost", by_exe="explorer.exe")
        popup_host = Windows.filter(by_class_name="ThumbnailDeviceHelperWnd", by_exe="explorer.exe")
        if popup_host:
            print(popup_host[0])
            popup_host[0].minimize()
            popup_host[0].close()

    windows = Windows.filter(by_exe="explorer.exe")
    while True:
        windows_new = Windows.filter(by_exe="explorer.exe")
        [print(w) for w in windows_new if w.hwnd not in [x.hwnd for x in windows]]
        windows = windows_new
        print(len(Windows.filter(by_exe="explorer.exe")))
        time.sleep(0.3)

    [print(w) for w in windows]
    w = windows[0]
    if w.is_uwp():
        cw = w.uwp_child
        cw.minimize()

    w.minimize()
    print()



import sys
from pathlib import Path
from time import sleep

sys.path.append(str(Path(__file__).parent.parent.absolute()))

from src.winytils.windows import Window, Windows
from src.winytils.workstation import Workstation


def test_windows():
    windows = Windows.filter(by_exe="explorer.exe")
    [print(w) for w in windows]


def test2():
    sleep(3)
    window = Windows.foregrond()

    if window:
        if window.is_uwp():
            child = window.uwp_child
            if child:
                window = child
        print(window)
        window.icon.save("icon.png")


def test3():
    workstation = Workstation()
    workstation.lock()


if __name__ == "__main__":
    test3()



import ctypes, sys
import win32api, win32con

from pathlib import Path

ROOT = Path(__file__).parent.absolute()


def is_admin():
    try:
        return ctypes.windll.shell32.IsUserAnAdmin()
    except:
        return False


def rerun_with_admin_privileges():
    # Re-run the program with admin rights
    # ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, " ".join(sys.argv), None, 1)
    win32api.ShellExecute(None, "runas", sys.executable, " ".join(sys.argv), None, win32con.SW_SHOWNORMAL)


if __name__ == "__main__":
    """
    https://stackoverflow.com/questions/130763/request-uac-elevation-from-within-a-python-script
    Also note that if you converted you python script into an executable file (using tools like py2exe, cx_freeze, pyinstaller) then you should use sys.argv[1:] instead of sys.argv in the fourth parameter.
    """

    if is_admin():
        with open(ROOT / "result.txt", "w", encoding="utf-8") as f:
            f.write("admin")
    else:
        rerun_with_admin_privileges()



import os
import ctypes


class Workstation:
    def __init__(self):
        pass

    def shutdown(self, *, force=False, delay: int = None):
        if force:
            os.system(f"shutdown /s /f /t 0") # does force apps close
            # os.system(f"shutdown /p") # Does not force close apps but instant shutdown
        elif delay:
            os.system(f"shutdown /s /t {delay}")
        else:
            # Wait for apps close and delay 30 sec
            os.system(f"shutdown /s")

    def restart(self, *, force=False, delay: int = None):
        if force:
            os.system(f"shutdown /r /f /t 0")
        elif delay:
            os.system(f"shutdown /r /t {delay}")
        else:
            # Wait for apps close and delay 30 sec
            os.system(f"shutdown /r")

    def hibernate(self):
        os.system(f"shutdown /h")

    def log_out(self):
        os.system(f"shutdown /l")

    def lock(self):
        ctypes.windll.user32.LockWorkStation()




"""
To execute every module directly - remove relative imports "."
"""

from . import machine_state_watcher
from . import privileges
from . import windows
from . import workstation


__version__ = "0.0.2"
